from PyCompGeomAlgorithms.core import Point
from AlgoGrade.parsers import PointListAndTargetPointGivenJSONParser


def test_point_list_parser():
    points = [Point(1, 1), Point(2, 2)]
    target_point = Point(3, 3)
    data = [[{"coords": [1, 1]}, {"coords": [2, 2]}], {"coords": [3, 3]}]

    assert PointListAndTargetPointGivenJSONParser.parse(data) == (points, target_point)


from __future__ import annotations
from functools import partial
from typing import ClassVar, Optional
from PyCompGeomAlgorithms.core import PathDirection
from PyCompGeomAlgorithms.preparata import preparata, PreparataNode, PreparataThreadedBinTree
from .core import Task, Grader, Answers
from .adapters import PointPydanticAdapter, ThreadedBinTreeNodePydanticAdapter, ThreadedBinTreePydanticAdapter
from .parsers import PointListGivenJSONParser


class PreparataGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [
            cls.grade_iterable,
            # ((left_paths, left_supporting_points), (right_paths, right_supporting_points)) -> paths is a list of lists of directions, supporting points is a list of points.
            partial(cls.grade_iterable, grade_item_method=partial(cls.grade_iterable, grade_item_method=[partial(cls.grade_iterable, grade_item_method=cls.grade_iterable), cls.grade_iterable])),
            partial(cls.grade_iterable, grade_item_method=cls.grade_iterable),
            partial(cls.grade_iterable, grade_item_method=(cls.grade_iterable, partial(cls.grade_iterable, grade_item_method=cls.grade_bin_tree)))
        ]


class PreparataNodePydanticAdapter(ThreadedBinTreeNodePydanticAdapter):
    regular_class: ClassVar[type] = PreparataNode
    data: PointPydanticAdapter
    left: Optional[PreparataNodePydanticAdapter] = None
    right: Optional[PreparataNodePydanticAdapter] = None

    @classmethod
    def from_regular_object(cls, obj: PreparataNode, **kwargs):
        return super().from_regular_object(obj, **kwargs)


class PreparataThreadedBinTreePydanticAdapter(ThreadedBinTreePydanticAdapter):
    regular_class: ClassVar[type] = PreparataThreadedBinTree
    root: PreparataNodePydanticAdapter

    @classmethod
    def from_regular_object(cls, obj: PreparataThreadedBinTree, **kwargs):
        return super().from_regular_object(obj, **kwargs)


class PreparataAnswers(Answers):
    hull: list[PointPydanticAdapter]
    tree: PreparataThreadedBinTreePydanticAdapter
    left_paths: list[list[PathDirection]]
    right_paths: list[list[PathDirection]]
    left_supporting_points: list[PointPydanticAdapter]
    right_supporting_points: list[PointPydanticAdapter]
    deleted_points_lists: list[list[PointPydanticAdapter]]
    hulls: list[list[PointPydanticAdapter]]
    trees: list[PreparataThreadedBinTreePydanticAdapter]

    @classmethod
    def from_iterable(cls, iterable):
        (hull, tree), ((left_paths, left_supporting_points), (right_paths, right_supporting_points)), deleted_points_lists, (hulls, trees), *rest = iterable
        return cls(
            hull=hull, tree=tree, left_paths=left_paths, right_paths=right_paths,
            left_supporting_points=left_supporting_points, right_supporting_points=right_supporting_points,
            deleted_points_lists=deleted_points_lists, hulls=hulls, trees=trees
        )
    
    def to_pydantic_list(self):
        return [
            (self.hull, self.tree),
            ((self.left_paths, self.left_supporting_points), (self.right_paths, self.right_supporting_points)),
            self.deleted_points_lists,
            (self.hulls, self.trees)
        ]


class PreparataTask(Task):
    algorithm = preparata
    grader_class = PreparataGrader
    answers_class = PreparataAnswers
    given_parser_class = PointListGivenJSONParser



from pytest import fixture
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.dynamic_hull import DynamicHullNode, DynamicHullTree, SubhullNode, SubhullThreadedBinTree
from AlgoGrade.adapters import PointPydanticAdapter, BinTreeNodePydanticAdapter
from AlgoGrade.dynamic_hull import DynamicHullNodePydanticAdapter, DynamicHullTreePydanticAdapter, SubhullNodePydanticAdapter, SubhullThreadedBinTreePydanticAdapter

@fixture
def dynamic_hull_node_adapter():
    point_adapter = PointPydanticAdapter(coords=(1, 1))
    return DynamicHullNodePydanticAdapter(
        data=point_adapter,
        subhull_points=[point_adapter],
        optimized_subhull_points=[point_adapter],
        left_supporting=point_adapter,
        right_supporting=point_adapter
    )


@fixture
def dynamic_hull_node_regular():
    point = Point(1, 1)
    return DynamicHullNode(point, [point], optimized_subhull_points=[point])


@fixture
def subhull_node_adapter():
    return SubhullNodePydanticAdapter(
        data=PointPydanticAdapter(coords=(1, 1))
    )


@fixture
def subhull_node_regular():
    return SubhullNode(Point(1, 1))


def test_dynamic_hull_node_adapter(dynamic_hull_node_adapter, dynamic_hull_node_regular):
    assert dynamic_hull_node_adapter.regular_object() == dynamic_hull_node_regular
    assert DynamicHullNodePydanticAdapter.from_regular_object(dynamic_hull_node_regular) == dynamic_hull_node_adapter


def test_dynamic_hull_node_adapter_serialization(dynamic_hull_node_adapter):
    serialized_node = dynamic_hull_node_adapter.model_dump()
    deserialized_node = DynamicHullNodePydanticAdapter(**serialized_node)
    assert deserialized_node.regular_object() == dynamic_hull_node_adapter.regular_object()


def test_dynamic_hull_tree_adapter(dynamic_hull_node_adapter, dynamic_hull_node_regular):
    tree_adapter = DynamicHullTreePydanticAdapter(root=dynamic_hull_node_adapter)
    regular_object = DynamicHullTree(dynamic_hull_node_regular)

    assert tree_adapter.regular_object() == regular_object
    assert DynamicHullTreePydanticAdapter.from_regular_object(regular_object) == tree_adapter


def test_dynamic_hull_tree_adapter_serialization(dynamic_hull_node_adapter):
    tree_adapter = DynamicHullTreePydanticAdapter(root=dynamic_hull_node_adapter)
    serialized_tree = tree_adapter.model_dump()
    deserialized_tree = DynamicHullTreePydanticAdapter(**serialized_tree)
    assert deserialized_tree.regular_object().root.data == tree_adapter.regular_object().root.data


def test_subhull_node_adapter(subhull_node_adapter, subhull_node_regular):
    assert subhull_node_adapter.regular_object() == subhull_node_regular
    assert SubhullNodePydanticAdapter.from_regular_object(subhull_node_regular) == subhull_node_adapter


def test_subhull_node_adapter_serialization(subhull_node_adapter):
    serialized_node = subhull_node_adapter.model_dump()
    deserialized_node = SubhullNodePydanticAdapter(**serialized_node)
    assert deserialized_node.regular_object() == subhull_node_adapter.regular_object()


def test_subhull_tree_adapter(subhull_node_adapter, subhull_node_regular):
    tree_adapter = SubhullThreadedBinTreePydanticAdapter(root=subhull_node_adapter)
    regular_object = SubhullThreadedBinTree(root=subhull_node_regular)

    assert tree_adapter.regular_object() == regular_object
    assert SubhullThreadedBinTreePydanticAdapter.from_regular_object(regular_object) == tree_adapter


def test_subhull_tree_adapter_serialization(subhull_node_adapter):
    tree_adapter = SubhullThreadedBinTreePydanticAdapter(root=subhull_node_adapter)
    serialized_tree = tree_adapter.model_dump()
    deserialized_tree = SubhullThreadedBinTreePydanticAdapter(**serialized_tree)
    assert deserialized_tree.regular_object() == tree_adapter.regular_object()


from collections import Counter
from typing import Iterable, Callable, Type, Any
from pydantic import BaseModel
from .adapters import pycga_to_pydantic, pydantic_to_pycga, SerializablePydanticModelWithPydanticFields


class Answers(SerializablePydanticModelWithPydanticFields):
    @classmethod
    def from_iterable(cls, iterable):
        raise NotImplementedError
    
    def to_pydantic_list(self):
        raise NotImplementedError
    
    def to_pycga_list(self):
        return pydantic_to_pycga(self.to_pydantic_list())


class Mistake:
    def __init__(self, scorings, description=""):
        self.scorings = scorings
        self.description = description
    
    @property
    def is_repeated(self):
        return self.scorings.repeat_fine > 0
    
    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.scorings == other.scorings
        )
    
    def __hash__(self):
        return hash((self.__class__, self.scorings))


class Scoring(BaseModel, frozen=True):
    min_grade: float = -1000.0
    max_grade: float = 0.0
    fine: float = 0.0
    repeat_fine: float = 0.0


class Grader:
    @classmethod
    def grade_pycga(cls, answers, correct_answers, scorings):
        """
            Compare answers to correct_answers (both in PyCGA format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PyCGA format.
        """
        mistake_lists = [
            grading_method(answer, correct_answer, scorings)
            for grading_method, answer, correct_answer, scorings
            in zip(cls.grade_methods(), answers, correct_answers, scorings)
        ]
        mistake_counters = [Counter(mistakes) for mistakes in mistake_lists]
        mistake_fines_dicts = [
            {
                mistake: mistake.scorings.repeat_fine if mistake.is_repeated and count > 1 else mistake.scorings.fine
                for mistake, count in mistake_counter.items()
            }
            for mistake_counter in mistake_counters
        ]

        answers_grades = [
            (answer, max(scorings.min_grade, scorings.max_grade-sum(mistake_fine_dict.values())))
            for answer, scorings, mistake_fine_dict
            in zip(answers, scorings, mistake_fines_dicts)
        ]
        total_grade = sum(grade for answer, grade in answers_grades)
        
        return total_grade, answers_grades
    
    @classmethod
    def grade_pydantic(cls, answers, correct_answers, scorings):
        """
            Compare answers to correct_answers (both in PydanticAdapter format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PydanticAdapter format.
        """
        total_grade, answers_grades = cls.grade_pycga(pydantic_to_pycga(answers), pydantic_to_pycga(correct_answers), scorings)
        return total_grade, [(pycga_to_pydantic(answer), grade) for answer, grade in answers_grades]
    
    @classmethod
    def grade_answers_wrapper(cls, answers: Answers, correct_answers: Answers, scorings):
        """
            Compare answers to correct_answers (both in Answers format) an return a tuple
            `(total_grade, answers_grades)`, where answers_grades is a list of tuples--grades for each answer `(answer, grade)` in PydanticAdapter format.
        """
        return cls.grade_pydantic(answers.to_pydantic_list(), correct_answers.to_pydantic_list(), scorings)
    
    @classmethod
    def grade_default(cls, answer, correct_answer, scorings, item_mistake_text=""):
        return [] if answer == correct_answer else [Mistake(scorings, description=item_mistake_text)]
    
    @classmethod
    def grade_iterable(cls, answer, correct_answer, scorings, grade_item_method=None, item_mistake_text=""):
        if grade_item_method is None:
            grade_item_method = cls.grade_default
        if callable(grade_item_method):
            grade_item_method = [grade_item_method] * len(correct_answer)
        elif not isinstance(grade_item_method, Iterable):
            raise TypeError(f"grade_item_method should be either a grading method, or an iterable of grading methods for each item, or None (defaults to default_grading)")

        len_diff = len(correct_answer) - len(answer)
        if len_diff == 0:
            return flatten([
                g(a, c, scorings, item_mistake_text=item_mistake_text)
                for g, a, c in zip(grade_item_method, answer, correct_answer)
            ])

        return [Mistake(scorings, description=f"Too {'few' if len_diff < 0 else 'many'} items")] * abs(len_diff)

    @classmethod
    def grade_bin_tree(cls, answer, correct_answer, scorings, grade_item_method=None, item_mistake_text=""):
        if grade_item_method is None:
            grade_item_method = cls.grade_default
        
        return cls._find_mistakes_in_bin_tree(answer.root, correct_answer.root, scorings, grade_item_method, item_mistake_text)

    @classmethod
    def _find_mistakes_in_bin_tree(
        cls, node, correct_node, scorings,
        grade_item_method=None, item_mistake_text="", mistakes=None, extra=None, missing=None
    ):
        if mistakes is None:
            mistakes = []
        if extra is None:
            extra = []
        if missing is None:
            missing = []
        
        mistakes.extend(grade_item_method(node, correct_node, scorings))
        
        if node.left and correct_node.left:
            cls._find_mistakes_in_bin_tree(node.left, correct_node.left, scorings, grade_item_method, item_mistake_text, mistakes, extra, missing)
        if node.left and not correct_node.left:
            extra.extend(Mistake(scorings, "Extra item") for _ in node.left.traverse_preorder()[1:])
        if not node.left and correct_node.left:
            missing.extend(Mistake(scorings, "Missing item") for _ in correct_node.left.traverse_preorder()[1:])

        if node.right and correct_node.right:
            cls._find_mistakes_in_bin_tree(node.right, correct_node.right, scorings, grade_item_method, item_mistake_text, mistakes, extra, missing)
        if node.right and not correct_node.right:
            extra.extend(Mistake(scorings, "Extra item") for _ in node.right.traverse_preorder()[1:])
        if not node.right and correct_node.right:
            missing.extend(Mistake(scorings, "Missing item") for _ in correct_node.right.traverse_preorder()[1:])
        
        return mistakes + extra + missing

    @classmethod
    def grade_methods(cls):
        raise NotImplementedError


class GivenJSONParser:
    """Parses a JSON-represented (list) givens as a tuple of arguments to be fed into the respective PyCGA task solving method."""
    @classmethod
    def parse(cls, data) -> list[Any]:
        """Parses a JSON-represented (list) givens as a tuple of arguments to be fed into the respective PyCGA task solving method."""
        raise NotImplementedError


class Task:
    algorithm: Callable = None
    grader_class: Type[Grader] = None
    answers_class: Type[Answers] = None
    given_parser_class: Type[GivenJSONParser] = None
    
    @classmethod
    def solve_as_pycga_list(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns a list of answers in PyCGA format."""
        return list(cls.algorithm(*givens))
    
    @classmethod
    def solve_as_pydantic_list(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns a list of answers in PydanticAdapter format."""
        pycga_answers = list(cls.solve_as_pycga_list(givens))
        return pycga_to_pydantic(pycga_answers)
    
    @classmethod
    def solve_as_answers_wrapper(cls, givens: Iterable):
        """Solves a task with the givens provided in PyCGA format and returns answers in form of the respective Answers wrapper."""
        pydantic_answers = cls.solve_as_pydantic_list(givens)
        return cls.answers_class.from_iterable(pydantic_answers)


def flatten(iterable):
    return list(_flatten(iterable))


def _flatten(iterable):
    for item in iterable:
        if isinstance(item, Iterable) and not isinstance(item, str):
            for inner_item in flatten(item):
                yield inner_item
        else:
            yield item



from AlgoGrade.adapters import PointPydanticAdapter, pydantic_to_pycga
from AlgoGrade.quickhull import QuickhullAnswers, QuickhullNodePydanticAdapter, QuickhullTreePydanticAdapter


def test_quickhull_answers():
    point = PointPydanticAdapter(coords=(1, 1))
    leftmost_point, rightmost_point = point, point
    subset1, subset2 = [point], [point]
    tree = QuickhullTreePydanticAdapter(root=QuickhullNodePydanticAdapter(data=[point]))

    answers_model = QuickhullAnswers(
        leftmost_point=leftmost_point, rightmost_point=rightmost_point,
        subset1=subset1, subset2=subset2, tree=tree
    )
    answers_list = [(leftmost_point, rightmost_point, subset1, subset2), tree, tree, tree, tree]

    assert answers_model.to_pydantic_list() == answers_list
    assert answers_model.to_pycga_list() == pydantic_to_pycga(answers_list)
    assert QuickhullAnswers.from_iterable(answers_list) == answers_model
    assert QuickhullAnswers(**answers_model.model_dump()) == answers_model



import json
from PyCompGeomAlgorithms.core import Point
from .core import GivenJSONParser


class PointListGivenJSONParser(GivenJSONParser):
    @classmethod
    def parse(cls, data):
        return ([Point(*item['coords']) for item in data],)


class PointListAndTargetPointGivenJSONParser(Point):
    @classmethod
    def parse(cls, data):
        points, target_point = data
        return [Point(*item['coords']) for item in points], Point(*target_point['coords'])


from setuptools import setup, find_packages


setup(
    name="AlgoGrade",
    version="1.0.14",
    author="artandfi (Artem Fisunenko)",
    author_email="artyom.fisunenko@gmail.com",
    description="A library for automated grading of algorithmic tasks with grading of their intermediate stages. "
    "The grading of some computational geometry tasks are provided out of the box.",
    packages=find_packages(),
    install_requires=[line.strip() for line in open("requirements.txt", "r").readlines()],
    keywords=[
        "Python3",
        "automated grading",
        "computational geometry",
        "convex hull",
        "region search",
        "geometric search",
        "point location",
        "proximity",
        "closest pair",
        "closest points"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ]
)



from PyCompGeomAlgorithms.core import PathDirection
from AlgoGrade.adapters import PointPydanticAdapter, pydantic_to_pycga
from AlgoGrade.preparata import PreparataNodePydanticAdapter, PreparataThreadedBinTreePydanticAdapter, PreparataAnswers


def test_preparata_answers():
    point = PointPydanticAdapter(coords=(1, 1))
    hull = [point]
    tree = PreparataThreadedBinTreePydanticAdapter(root=PreparataNodePydanticAdapter(data=point))
    left_paths, right_paths = [[PathDirection.left]], [[PathDirection.right]]
    left_supporting_points, right_supporting_points = [point], [point]
    deleted_points = []
    hulls, trees = [hull], [tree]

    answers_model = PreparataAnswers(
        hull=hull, tree=tree, left_paths=left_paths, right_paths=right_paths,
        left_supporting_points=left_supporting_points, right_supporting_points=right_supporting_points,
        deleted_points_lists=deleted_points, hulls=hulls, trees=trees
    )
    answers_list = [(hull, tree), ((left_paths, left_supporting_points), (right_paths, right_supporting_points)), deleted_points, (hulls, trees)]

    assert answers_model.to_pydantic_list() == answers_list
    assert answers_model.to_pycga_list() == pydantic_to_pycga(answers_list)
    assert PreparataAnswers.from_iterable(answers_list) == answers_model
    assert PreparataAnswers(**answers_model.model_dump()) == answers_model
